@@ -10,11 +10,13 @@ use std::str;
 
 use anyhow::{bail, Context, Result};
 use fs_err as fs;
-use target_lexicon::{OperatingSystem, Triple};
+#[cfg(not(target_family = "unix"))]
+use path_slash::PathBufExt;
+use serde::Deserialize;
+use target_lexicon::{Architecture, Environment, OperatingSystem, Triple};
 
 /// Zig linker wrapper
-#[derive(Debug, clap::Subcommand)]
-#[clap(hide = true)]
+#[derive(Clone, Debug, clap::Subcommand)]
 pub enum Zig {
     /// `zig cc` wrapper
     #[clap(name = "cc", trailing_var_arg = true)]
@@ -39,6 +41,7 @@ impl Zig {
             Zig::Cc { args } => ("cc", args),
             Zig::Cxx { args } => ("c++", args),
         };
+
         let target = cmd_args
             .iter()
             .position(|x| x == "-target")
@@ -47,12 +50,19 @@ impl Zig {
         let is_windows_gnu = target
             .map(|x| x.contains("windows-gnu"))
             .unwrap_or_default();
+        let is_arm = target.map(|x| x.contains("arm")).unwrap_or_default();
+
+        let rustc_ver = rustc_version::version()?;
 
         let filter_link_arg = |arg: &str| {
             if arg == "-lgcc_s" {
                 // Replace libgcc_s with libunwind
                 return Some("-lunwind".to_string());
             }
+            if is_arm && arg.ends_with(".rlib") && arg.contains("libcompiler_builtins-") {
+                // compiler-builtins is duplicated with zig's compiler-rt
+                return None;
+            }
             if is_windows_gnu {
                 if arg == "-lgcc_eh" {
                     // zig doesn't provide gcc_eh alternative
@@ -67,7 +77,13 @@ impl Zig {
                 if arg.ends_with(".o") && arg.contains("self-contained") && arg.contains("crt") {
                     return None;
                 }
-                if arg.ends_with(".rlib") && arg.contains("liblibc-") {
+                if rustc_ver.major == 1
+                    && rustc_ver.minor < 59
+                    && arg.ends_with(".rlib")
+                    && arg.contains("liblibc-")
+                {
+                    // Rust distributes standalone libc.a in self-contained for musl since 1.59.0
+                    // See https://github.com/rust-lang/rust/pull/90527
                     return None;
                 }
                 if arg == "-lc" {
@@ -81,10 +97,7 @@ impl Zig {
                 .iter()
                 .position(|x| x == "-undefined")
                 .and_then(|i| args.get(i + 1));
-            match undefined {
-                Some(x) if x == "dynamic_lookup" => true,
-                _ => false,
-            }
+            matches!(undefined, Some(x) if x == "dynamic_lookup")
         };
 
         let mut new_cmd_args = Vec::with_capacity(cmd_args.len());
@@ -109,7 +122,7 @@ impl Zig {
                 new_cmd_args.push(arg);
             }
         }
-        if has_undefined_dynamic_lookup(&cmd_args) {
+        if has_undefined_dynamic_lookup(cmd_args) {
             new_cmd_args.push("-Wl,-undefined=dynamic_lookup".to_string());
         }
         let (zig, zig_args) = Self::find_zig()?;
@@ -169,6 +182,19 @@ impl Zig {
             )
         }
     }
+
+    /// Find zig lib directory
+    pub fn lib_dir() -> Result<PathBuf> {
+        let (zig, zig_args) = Self::find_zig()?;
+        let output = Command::new(zig).args(zig_args).arg("env").output()?;
+        let zig_env: ZigEnv = serde_json::from_slice(&output.stdout)?;
+        Ok(PathBuf::from(zig_env.lib_dir))
+    }
+}
+
+#[derive(Debug, Deserialize)]
+struct ZigEnv {
+    lib_dir: String,
 }
 
 /// Prepare wrapper scripts for `zig cc` and `zig c++` and returns their paths
@@ -200,22 +226,39 @@ pub fn prepare_zig_linker(target: &str) -> Result<(PathBuf, PathBuf)> {
         format!(".{}", abi_suffix)
     };
     let triple: Triple = rust_target.parse().unwrap();
-    let arch = triple.architecture.to_string();
-    let file_ext = if cfg!(windows) { "bat" } else { "sh" };
+    let arch = match triple.architecture {
+        // zig target only has i386, no i586/i686
+        Architecture::X86_32(..) => "i386".to_string(),
+        architecture => architecture.to_string(),
+    };
+    let target_env = match (triple.architecture, triple.environment) {
+        (Architecture::Mips32(..), Environment::Gnu) => Environment::Gnueabihf,
+        (_, environment) => environment,
+    };
+    let file_ext = if cfg!(windows) {
+        if is_mingw_shell() {
+            // mingw or msys2
+            "sh"
+        } else {
+            "bat"
+        }
+    } else {
+        "sh"
+    };
     let zig_cc = format!("zigcc-{}.{}", target, file_ext);
     let zig_cxx = format!("zigcxx-{}.{}", target, file_ext);
     let cc_args = "-g"; // prevent stripping
     let cc_args = match triple.operating_system {
         OperatingSystem::Linux => format!(
             "-target {}-linux-{}{} {}",
-            arch, triple.environment, abi_suffix, cc_args,
+            arch, target_env, abi_suffix, cc_args,
         ),
         OperatingSystem::MacOSX { .. } | OperatingSystem::Darwin => {
             format!("-target {}-macos-gnu{} {}", arch, abi_suffix, cc_args)
         }
         OperatingSystem::Windows { .. } => format!(
             "-target {}-windows-{}{} {}",
-            arch, triple.environment, abi_suffix, cc_args,
+            arch, target_env, abi_suffix, cc_args,
         ),
         _ => bail!("unsupported target"),
     };
@@ -269,12 +312,19 @@ fn write_linker_wrapper(path: &Path, command: &str, args: &str) -> Result<()> {
     } else {
         env::current_exe()?
     };
+    let current_exe = if is_mingw_shell() {
+        current_exe.to_slash_lossy()
+    } else {
+        current_exe.display().to_string()
+    };
     writeln!(
         &mut custom_linker_file,
         "{} zig {} -- {} %*",
-        current_exe.display(),
-        command,
-        args
+        current_exe, command, args
     )?;
     Ok(())
 }
+
+pub(crate) fn is_mingw_shell() -> bool {
+    env::var_os("MSYSTEM").is_some() && env::var_os("SHELL").is_some()
+}
